@@ -41,6 +41,9 @@ Coming in from initial call from client or customer
 HTTP_X_AUTH_TOKEN
     The client token being passed in.
 
+HTTP_X_SERVICE_TOKEN
+    A service token being passed in.
+
 HTTP_X_STORAGE_TOKEN
     The client token being passed in (legacy Rackspace use) to support
     swift/cloud files
@@ -55,55 +58,61 @@ WWW-Authenticate
 What we add to the request for use by the OpenStack service
 ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 
+When using composite authentication (a user and service token are
+present) additional service headers relating to the service user
+will be added. They take the same form as the standard headers but add
+'_SERVICE_'. These headers will not exist in the environment if no
+service token is present.
+
 HTTP_X_IDENTITY_STATUS
     'Confirmed' or 'Invalid'
     The underlying service will only see a value of 'Invalid' if the Middleware
     is configured to run in 'delay_auth_decision' mode
 
-HTTP_X_DOMAIN_ID
+HTTP_X_DOMAIN_ID, HTTP_X_SERVICE_DOMAIN_ID
     Identity service managed unique identifier, string. Only present if
     this is a domain-scoped v3 token.
 
-HTTP_X_DOMAIN_NAME
+HTTP_X_DOMAIN_NAME, HTTP_X_SERVICE_DOMAIN_NAME
     Unique domain name, string. Only present if this is a domain-scoped
     v3 token.
 
-HTTP_X_PROJECT_ID
+HTTP_X_PROJECT_ID, HTTP_X_SERVICE_PROJECT_ID
     Identity service managed unique identifier, string. Only present if
     this is a project-scoped v3 token, or a tenant-scoped v2 token.
 
-HTTP_X_PROJECT_NAME
+HTTP_X_PROJECT_NAME, HTTP_X_SERVICE_PROJECT_NAME
     Project name, unique within owning domain, string. Only present if
     this is a project-scoped v3 token, or a tenant-scoped v2 token.
 
-HTTP_X_PROJECT_DOMAIN_ID
+HTTP_X_PROJECT_DOMAIN_ID, HTTP_X_SERVICE_PROJECT_DOMAIN_ID
     Identity service managed unique identifier of owning domain of
     project, string.  Only present if this is a project-scoped v3 token. If
     this variable is set, this indicates that the PROJECT_NAME can only
     be assumed to be unique within this domain.
 
-HTTP_X_PROJECT_DOMAIN_NAME
+HTTP_X_PROJECT_DOMAIN_NAME, HTTP_X_SERVICE_PROJECT_DOMAIN_NAME
     Name of owning domain of project, string. Only present if this is a
     project-scoped v3 token. If this variable is set, this indicates that
     the PROJECT_NAME can only be assumed to be unique within this domain.
 
-HTTP_X_USER_ID
+HTTP_X_USER_ID, HTTP_X_SERVICE_USER_ID
     Identity-service managed unique identifier, string
 
-HTTP_X_USER_NAME
+HTTP_X_USER_NAME, HTTP_X_SERVICE_USER_NAME
     User identifier, unique within owning domain, string
 
-HTTP_X_USER_DOMAIN_ID
+HTTP_X_USER_DOMAIN_ID, HTTP_X_SERVICE_USER_DOMAIN_ID
     Identity service managed unique identifier of owning domain of
     user, string. If this variable is set, this indicates that the USER_NAME
     can only be assumed to be unique within this domain.
 
-HTTP_X_USER_DOMAIN_NAME
+HTTP_X_USER_DOMAIN_NAME, HTTP_X_SERVICE_USER_DOMAIN_NAME
     Name of owning domain of user, string. If this variable is set, this
     indicates that the USER_NAME can only be assumed to be unique within
     this domain.
 
-HTTP_X_ROLES
+HTTP_X_ROLES, HTTP_X_SERVICE_ROLES
     Comma delimited list of case-sensitive role names
 
 HTTP_X_SERVICE_CATALOG
@@ -111,6 +120,11 @@ HTTP_X_SERVICE_CATALOG
     For compatibility reasons this catalog will always be in the V2 catalog
     format even if it is a v3 token.
 
+    Note: This is an exception in that it contains 'SERVICE' but relates to a
+          user token, not a service token. The existing user's
+          catalog can be very large; it was decided not to present a catalog
+          relating to the service token to avoid using more HTTP header space.
+
 HTTP_X_TENANT_ID
     *Deprecated* in favor of HTTP_X_PROJECT_ID
     Identity service managed unique identifier, string. For v3 tokens, this
@@ -143,6 +157,11 @@ keystone.token_info
     Keystone token validation call, as well as basic information about
     the tenant and user.
 
+keystone.token_auth
+    A keystoneclient auth plugin that may be used with a
+    :py:class:`keystoneclient.session.Session`. This plugin will load the
+    authentication data provided to auth_token middleware.
+
 """
 
 import contextlib
@@ -154,6 +173,7 @@ import tempfile
 import time
 
 from keystoneclient import access
+from keystoneclient.auth.identity import base as base_identity
 from keystoneclient.auth.identity import v2
 from keystoneclient.auth import token_endpoint
 from keystoneclient.common import cms
@@ -302,6 +322,31 @@ _OPTS = [
                secret=True,
                help='(optional, mandatory if memcache_security_strategy is'
                ' defined) this string is used for key derivation.'),
+    cfg.IntOpt('memcache_pool_dead_retry',
+               default=5 * 60,
+               help='(optional) number of seconds memcached server is'
+               ' considered dead before it is tried again.'),
+    cfg.IntOpt('memcache_pool_maxsize',
+               default=10,
+               help='(optional) max total number of open connections to'
+               ' every memcached server.'),
+    cfg.IntOpt('memcache_pool_socket_timeout',
+               default=3,
+               help='(optional) socket timeout in seconds for communicating '
+                    'with a memcache server.'),
+    cfg.IntOpt('memcache_pool_unused_timeout',
+               default=60,
+               help='(optional) number of seconds a connection to memcached'
+               ' is held unused in the pool before it is closed.'),
+    cfg.IntOpt('memcache_pool_conn_get_timeout',
+               default=10,
+               help='(optional) number of seconds that an operation will wait '
+                    'to get a memcache client connection from the pool.'),
+    cfg.BoolOpt('memcache_use_advanced_pool',
+                default=False,
+                help='(optional) use the advanced (eventlet safe) memcache '
+                     'client pool. The advanced pool will only work under '
+                     'python 2.x.'),
     cfg.BoolOpt('include_service_catalog',
                 default=True,
                 help='(optional) indicate whether to set the X-Service-Catalog'
@@ -339,6 +384,26 @@ CONF.register_opts(_OPTS, group='keystone_authtoken')
 
 _LIST_OF_VERSIONS_TO_ATTEMPT = ['v3.0', 'v2.0']
 
+_HEADER_TEMPLATE = {
+    'X%s-Domain-Id': 'domain_id',
+    'X%s-Domain-Name': 'domain_name',
+    'X%s-Project-Id': 'project_id',
+    'X%s-Project-Name': 'project_name',
+    'X%s-Project-Domain-Id': 'project_domain_id',
+    'X%s-Project-Domain-Name': 'project_domain_name',
+    'X%s-User-Id': 'user_id',
+    'X%s-User-Name': 'username',
+    'X%s-User-Domain-Id': 'user_domain_id',
+    'X%s-User-Domain-Name': 'user_domain_name',
+}
+
+_DEPRECATED_HEADER_TEMPLATE = {
+    'X-User': 'username',
+    'X-Tenant-Id': 'project_id',
+    'X-Tenant-Name': 'project_name',
+    'X-Tenant': 'project_name',
+}
+
 
 class _BIND_MODE:
     DISABLED = 'disabled'
@@ -366,20 +431,25 @@ def _token_is_v3(token_info):
     return ('token' in token_info)
 
 
-def _confirm_token_not_expired(data):
+def _get_token_expiration(data):
     if not data:
-        raise InvalidUserToken('Token authorization failed')
+        raise InvalidToken('Token authorization failed')
     if _token_is_v2(data):
         timestamp = data['access']['token']['expires']
     elif _token_is_v3(data):
         timestamp = data['token']['expires_at']
     else:
-        raise InvalidUserToken('Token authorization failed')
+        raise InvalidToken('Token authorization failed')
     expires = timeutils.parse_isotime(timestamp)
     expires = timeutils.normalize_time(expires)
+    return expires
+
+
+def _confirm_token_not_expired(data):
+    expires = _get_token_expiration(data)
     utcnow = timeutils.utcnow()
     if utcnow >= expires:
-        raise InvalidUserToken('Token authorization failed')
+        raise InvalidToken('Token authorization failed')
     return timeutils.isotime(at=expires, subsecond=True)
 
 
@@ -424,7 +494,28 @@ def _safe_quote(s):
     return urllib.parse.quote(s) if s == urllib.parse.unquote(s) else s
 
 
-class InvalidUserToken(Exception):
+def _conf_values_type_convert(conf):
+    """Convert conf values into correct type."""
+    if not conf:
+        return {}
+    opts = {}
+    opt_types = dict((o.dest, getattr(o, 'type', str)) for o in _OPTS)
+    for k, v in six.iteritems(conf):
+        try:
+            if v is None:
+                opts[k] = v
+            else:
+                opts[k] = opt_types[k](v)
+        except KeyError:
+            opts[k] = v
+        except ValueError as e:
+            raise ConfigurationError(
+                'Unable to convert the value of %s option into correct '
+                'type: %s' % (k, e))
+    return opts
+
+
+class InvalidToken(Exception):
     pass
 
 
@@ -453,13 +544,48 @@ class _MiniResp(object):
         self.headers.append(('Content-type', 'text/plain'))
 
 
+class _UserAuthPlugin(base_identity.BaseIdentityPlugin):
+    """The incoming authentication credentials.
+
+    A plugin that represents the incoming user credentials. This can be
+    consumed by applications.
+
+    This object is not expected to be constructed directly by users. It is
+    created and passed by auth_token middleware and then can be used as the
+    authentication plugin when communicating via a session.
+    """
+
+    def __init__(self, user_token, auth_ref):
+        # FIXME(jamielennox): set reauthenticate=False here when keystoneclient
+        # 0.11 is released to prevent trying to refetch authentication.
+        super(_UserAuthPlugin, self).__init__()
+        self._user_token = user_token
+        self._stored_auth_ref = auth_ref
+
+    def get_token(self, session, **kwargs):
+        # NOTE(jamielennox): This is needed partially because the AccessInfo
+        # factory is so bad that we don't always get the correct token data.
+        # Override and always return the token that was provided in the req.
+        return self._user_token
+
+    def get_auth_ref(self, session, **kwargs):
+        # NOTE(jamielennox): We can't go out and fetch this auth_ref, we've
+        # got it already so always return it. In the event it tries to
+        # re-authenticate it will get the same old auth_ref which is not
+        # perfect, but the best we can do for now.
+        return self._stored_auth_ref
+
+
 class AuthProtocol(object):
     """Auth Middleware that handles authenticating client calls."""
 
     def __init__(self, app, conf):
         self._LOG = logging.getLogger(conf.get('log_name', __name__))
         self._LOG.info('Starting keystone auth_token middleware')
-        self._conf = conf
+        # NOTE(wanghong): If options are set in paste file, all the option
+        # values passed into conf are string type. So, we should convert the
+        # conf value into correct type.
+        self._conf = _conf_values_type_convert(conf)
         self._app = app
 
         # delay_auth_decision means we still allow unauthenticated requests
@@ -536,6 +662,7 @@ class AuthProtocol(object):
 
         self._check_revocations_for_cached = self._conf_get(
             'check_revocations_for_cached')
+        self._init_auth_headers()
 
     def _conf_get(self, name):
         # try config from paste-deploy first
@@ -544,6 +671,18 @@ class AuthProtocol(object):
         else:
             return CONF.keystone_authtoken[name]
 
+    def _call_app(self, env, start_response):
+        # NOTE(jamielennox): We wrap the given start response so that if an
+        # application with a 'delay_auth_decision' setting fails, or otherwise
+        # raises Unauthorized that we include the Authentication URL headers.
+        def _fake_start_response(status, response_headers, exc_info=None):
+            if status.startswith('401'):
+                response_headers.extend(self._reject_auth_headers)
+
+            return start_response(status, response_headers, exc_info)
+
+        return self._app(env, _fake_start_response)
+
     def __call__(self, env, start_response):
         """Handle incoming request.
 
@@ -551,27 +690,51 @@ class AuthProtocol(object):
         we can't authenticate.
 
         """
-        self._LOG.debug('Authenticating user token')
+        def _fmt_msg(env):
+            msg = ('user: user_id %s, project_id %s, roles %s '
+                   'service: user_id %s, project_id %s, roles %s' % (
+                   env.get('X_USER_ID'), env.get('X_PROJECT_ID'),
+                   env.get('X_ROLES'), env.get('X_SERVICE_USER_ID'),
+                   env.get('X_SERVICE_PROJECT_ID'),
+                   env.get('X_SERVICE_ROLES')))
+            return msg
 
         self._token_cache.initialize(env)
+        self._remove_auth_headers(env)
 
         try:
-            self._remove_auth_headers(env)
-            user_token = self._get_user_token_from_header(env)
-            token_info = self._validate_user_token(user_token, env)
-            env['keystone.token_info'] = token_info
-            user_headers = self._build_user_headers(token_info)
-            self._add_headers(env, user_headers)
-            return self._app(env, start_response)
 
-        except InvalidUserToken:
-            if self._delay_auth_decision:
-                self._LOG.info(
-                    'Invalid user token - deferring reject downstream')
-                self._add_headers(env, {'X-Identity-Status': 'Invalid'})
-                return self._app(env, start_response)
-            else:
-                self._LOG.info('Invalid user token - rejecting request')
+            try:
+                self._LOG.debug('Authenticating user token')
+                user_token = self._get_user_token_from_header(env)
+                token_info = self._validate_token(user_token, env)
+                auth_ref = access.AccessInfo.factory(body=token_info)
+                env['keystone.token_info'] = token_info
+                env['keystone.token_auth'] = _UserAuthPlugin(
+                    user_token, auth_ref)
+                user_headers = self._build_user_headers(auth_ref, token_info)
+                self._add_headers(env, user_headers)
+            except InvalidToken:
+                if self._delay_auth_decision:
+                    self._LOG.info(
+                        'Invalid user token - deferring reject downstream')
+                    self._add_headers(env, {'X-Identity-Status': 'Invalid'})
+                else:
+                    self._LOG.info('Invalid user token - rejecting request')
+                    return self._reject_request(env, start_response)
+
+            try:
+                self._LOG.debug('Authenticating service token')
+                serv_token = self._get_service_token_from_header(env)
+                if serv_token is not None:
+                    serv_token_info = self._validate_token(
+                        serv_token, env)
+                    serv_headers = self._build_service_headers(serv_token_info)
+                    self._add_headers(env, serv_headers)
+            except InvalidToken:
+                # Delayed auth not currently supported for service tokens.
+                # (Can be implemented if a use case is found.)
+                self._LOG.info('Invalid service token - rejecting request')
                 return self._reject_request(env, start_response)
 
         except ServiceError as e:
@@ -580,43 +743,49 @@ class AuthProtocol(object):
             start_response('503 Service Unavailable', resp.headers)
             return resp.body
 
+        self._LOG.debug("Received request from %s" % _fmt_msg(env))
+
+        return self._call_app(env, start_response)
+
+    def _init_auth_headers(self):
+        """Initialize auth header list.
+
+        Both user and service token headers are generated.
+        """
+        auth_headers = ['X-Service-Catalog',
+                        'X-Identity-Status',
+                        'X-Roles',
+                        'X-Service-Roles']
+        for key in six.iterkeys(_HEADER_TEMPLATE):
+            auth_headers.append(key % '')
+            # Service headers
+            auth_headers.append(key % '-Service')
+
+        # Deprecated headers
+        auth_headers.append('X-Role')
+        for key in six.iterkeys(_DEPRECATED_HEADER_TEMPLATE):
+            auth_headers.append(key)
+
+        self._auth_headers = auth_headers
+
     def _remove_auth_headers(self, env):
         """Remove headers so a user can't fake authentication.
 
+        Both user and service token headers are removed.
+
         :param env: wsgi request environment
 
         """
-        auth_headers = (
-            'X-Identity-Status',
-            'X-Domain-Id',
-            'X-Domain-Name',
-            'X-Project-Id',
-            'X-Project-Name',
-            'X-Project-Domain-Id',
-            'X-Project-Domain-Name',
-            'X-User-Id',
-            'X-User-Name',
-            'X-User-Domain-Id',
-            'X-User-Domain-Name',
-            'X-Roles',
-            'X-Service-Catalog',
-            # Deprecated
-            'X-User',
-            'X-Tenant-Id',
-            'X-Tenant-Name',
-            'X-Tenant',
-            'X-Role',
-        )
         self._LOG.debug('Removing headers from request environment: %s',
-                        ','.join(auth_headers))
-        self._remove_headers(env, auth_headers)
+                        ','.join(self._auth_headers))
+        self._remove_headers(env, self._auth_headers)
 
     def _get_user_token_from_header(self, env):
         """Get token id from request.
 
         :param env: wsgi request environment
         :return token id
-        :raises InvalidUserToken if no token is provided in request
+        :raises InvalidToken if no token is provided in request
 
         """
         token = self._get_header(env, 'X-Auth-Token',
@@ -628,7 +797,21 @@ class AuthProtocol(object):
                 self._LOG.warn('Unable to find authentication token'
                                ' in headers')
                 self._LOG.debug('Headers: %s', env)
-            raise InvalidUserToken('Unable to find token in headers')
+            raise InvalidToken('Unable to find token in headers')
+
+    def _get_service_token_from_header(self, env):
+        """Get service token id from request.
+
+        :param env: wsgi request environment
+        :return service token id or None if not present
+
+        """
+        return self._get_header(env, 'X-Service-Token')
+
+    @property
+    def _reject_auth_headers(self):
+        header_val = 'Keystone uri=\'%s\'' % self._identity_server.auth_uri
+        return [('WWW-Authenticate', header_val)]
 
     def _reject_request(self, env, start_response):
         """Redirect client to auth server.
@@ -638,28 +821,32 @@ class AuthProtocol(object):
         :returns HTTPUnauthorized http response
 
         """
-        header_val = 'Keystone uri=\'%s\'' % self._identity_server.auth_uri
-        headers = [('WWW-Authenticate', header_val)]
-        resp = _MiniResp('Authentication required', env, headers)
+        resp = _MiniResp('Authentication required',
+                         env, self._reject_auth_headers)
         start_response('401 Unauthorized', resp.headers)
         return resp.body
 
-    def _validate_user_token(self, user_token, env, retry=True):
+    def _validate_token(self, token, env, retry=True):
         """Authenticate user token
 
-        :param user_token: user's token id
+        :param token: token id
         :param retry: Ignored, as it is not longer relevant
         :return uncrypted body of the token if the token is valid
-        :raise InvalidUserToken if token is rejected
+        :raise InvalidToken if token is rejected
         :no longer raises ServiceError since it no longer makes RPC
 
         """
         token_id = None
 
         try:
-            token_ids, cached = self._token_cache.get(user_token)
+            token_ids, cached = self._token_cache.get(token)
             token_id = token_ids[0]
             if cached:
+                # Token was retrieved from the cache. In this case, there's no
+                # need to check that the token is expired because the cache
+                # fetch fails for an expired token. Also, there's no need to
+                # put the token in the cache because it's already in the cache.
+
                 data = cached
 
                 if self._check_revocations_for_cached:
@@ -671,71 +858,68 @@ class AuthProtocol(object):
                         if is_revoked:
                             self._LOG.debug(
                                 'Token is marked as having been revoked')
-                            raise InvalidUserToken(
+                            raise InvalidToken(
                                 'Token authorization failed')
-            elif cms.is_pkiz(user_token):
-                verified = self._verify_pkiz_token(user_token, token_ids)
-                data = jsonutils.loads(verified)
-            elif cms.is_asn1_token(user_token):
-                verified = self._verify_signed_token(user_token, token_ids)
-                data = jsonutils.loads(verified)
+                self._confirm_token_bind(data, env)
             else:
-                data = self._identity_server.verify_token(user_token, retry)
-            expires = _confirm_token_not_expired(data)
-            self._confirm_token_bind(data, env)
-            self._token_cache.store(token_id, data, expires)
+                # Token wasn't cached. In this case, the token needs to be
+                # checked that it's not expired, and also put in the cache.
+                if cms.is_pkiz(token):
+                    verified = self._verify_pkiz_token(token, token_ids)
+                    data = jsonutils.loads(verified)
+                    expires = _confirm_token_not_expired(data)
+                elif cms.is_asn1_token(token):
+                    verified = self._verify_signed_token(token, token_ids)
+                    data = jsonutils.loads(verified)
+                    expires = _confirm_token_not_expired(data)
+                else:
+                    data = self._identity_server.verify_token(token,
+                                                              retry)
+                    # No need to confirm token expiration here since
+                    # verify_token fails for expired tokens.
+                    expires = _get_token_expiration(data)
+                self._confirm_token_bind(data, env)
+                self._token_cache.store(token_id, data, expires)
             return data
         except NetworkError:
             self._LOG.debug('Token validation failure.', exc_info=True)
             self._LOG.warn('Authorization failed for token')
-            raise InvalidUserToken('Token authorization failed')
+            raise InvalidToken('Token authorization failed')
         except Exception:
             self._LOG.debug('Token validation failure.', exc_info=True)
             if token_id:
                 self._token_cache.store_invalid(token_id)
             self._LOG.warn('Authorization failed for token')
-            raise InvalidUserToken('Token authorization failed')
+            raise InvalidToken('Token authorization failed')
 
-    def _build_user_headers(self, token_info):
+    def _build_user_headers(self, auth_ref, token_info):
         """Convert token object into headers.
 
         Build headers that represent authenticated user - see main
         doc info at start of file for details of headers to be defined.
 
-        :param token_info: token object returned by keystone on authentication
-        :raise InvalidUserToken when unable to parse token object
+        :param token_info: token object returned by identity
+                           server on authentication
+        :raise InvalidToken: when unable to parse token object
 
         """
-        auth_ref = access.AccessInfo.factory(body=token_info)
         roles = ','.join(auth_ref.role_names)
 
         if _token_is_v2(token_info) and not auth_ref.project_id:
-            raise InvalidUserToken('Unable to determine tenancy.')
+            raise InvalidToken('Unable to determine tenancy.')
 
         rval = {
             'X-Identity-Status': 'Confirmed',
-            'X-Domain-Id': auth_ref.domain_id,
-            'X-Domain-Name': auth_ref.domain_name,
-            'X-Project-Id': auth_ref.project_id,
-            'X-Project-Name': auth_ref.project_name,
-            'X-Project-Domain-Id': auth_ref.project_domain_id,
-            'X-Project-Domain-Name': auth_ref.project_domain_name,
-            'X-User-Id': auth_ref.user_id,
-            'X-User-Name': auth_ref.username,
-            'X-User-Domain-Id': auth_ref.user_domain_id,
-            'X-User-Domain-Name': auth_ref.user_domain_name,
             'X-Roles': roles,
-            # Deprecated
-            'X-User': auth_ref.username,
-            'X-Tenant-Id': auth_ref.project_id,
-            'X-Tenant-Name': auth_ref.project_name,
-            'X-Tenant': auth_ref.project_name,
-            'X-Role': roles,
         }
 
-        self._LOG.debug('Received request from user: %s with project_id : %s'
-                        ' and roles: %s ',
-                        auth_ref.user_id, auth_ref.project_id, roles)
+        for header_tmplt, attr in six.iteritems(_HEADER_TEMPLATE):
+            rval[header_tmplt % ''] = getattr(auth_ref, attr)
+
+        # Deprecated headers
+        rval['X-Role'] = roles
+        for header_tmplt, attr in six.iteritems(_DEPRECATED_HEADER_TEMPLATE):
+            rval[header_tmplt] = getattr(auth_ref, attr)
 
         if self._include_service_catalog and auth_ref.has_service_catalog():
             catalog = auth_ref.service_catalog.get_data()
@@ -745,6 +929,33 @@ class AuthProtocol(object):
 
         return rval
 
+    def _build_service_headers(self, token_info):
+        """Convert token object into service headers.
+
+        Build headers that represent authenticated user - see main
+        doc info at start of file for details of headers to be defined.
+
+        :param token_info: token object returned by identity
+                           server on authentication
+        :raise InvalidToken: when unable to parse token object
+
+        """
+        auth_ref = access.AccessInfo.factory(body=token_info)
+
+        if _token_is_v2(token_info) and not auth_ref.project_id:
+            raise InvalidToken('Unable to determine service tenancy.')
+
+        roles = ','.join(auth_ref.role_names)
+        rval = {
+            'X-Service-Roles': roles,
+        }
+
+        header_type = '-Service'
+        for header_tmplt, attr in six.iteritems(_HEADER_TEMPLATE):
+            rval[header_tmplt % header_type] = getattr(auth_ref, attr)
+
+        return rval
+
     def _header_to_env_var(self, key):
         """Convert header to wsgi env variable.
 
@@ -779,7 +990,7 @@ class AuthProtocol(object):
         if msg is False:
             msg = 'Token authorization failed'
 
-        raise InvalidUserToken(msg)
+        raise InvalidToken(msg)
 
     def _confirm_token_bind(self, data, env):
         bind_mode = self._conf_get('enforce_token_bind')
@@ -897,7 +1108,7 @@ class AuthProtocol(object):
     def _verify_signed_token(self, signed_text, token_ids):
         """Check that the token is unrevoked and has a valid signature."""
         if self._is_signed_token_revoked(token_ids):
-            raise InvalidUserToken('Token has been revoked')
+            raise InvalidToken('Token has been revoked')
 
         formatted = cms.token_to_cms(signed_text)
         verified = self._cms_verify(formatted)
@@ -905,14 +1116,14 @@ class AuthProtocol(object):
 
     def _verify_pkiz_token(self, signed_text, token_ids):
         if self._is_signed_token_revoked(token_ids):
-            raise InvalidUserToken('Token has been revoked')
+            raise InvalidToken('Token has been revoked')
         try:
             uncompressed = cms.pkiz_uncompress(signed_text)
             verified = self._cms_verify(uncompressed, inform=cms.PKIZ_CMS_FORM)
             return verified
         # TypeError If the signed_text is not zlib compressed
         except TypeError:
-            raise InvalidUserToken(signed_text)
+            raise InvalidToken(signed_text)
 
     def _verify_signing_dir(self):
         if os.path.exists(self._signing_dirname):
@@ -1059,7 +1270,17 @@ class AuthProtocol(object):
             env_cache_name=self._conf_get('cache'),
             memcached_servers=self._conf_get('memcached_servers'),
             memcache_security_strategy=self._memcache_security_strategy,
-            memcache_secret_key=self._conf_get('memcache_secret_key'))
+            memcache_secret_key=self._conf_get('memcache_secret_key'),
+            use_advanced_pool=self._conf_get('memcache_use_advanced_pool'),
+            memcache_pool_dead_retry=self._conf_get(
+                'memcache_pool_dead_retry'),
+            memcache_pool_maxsize=self._conf_get('memcache_pool_maxsize'),
+            memcache_pool_unused_timeout=self._conf_get(
+                'memcache_pool_unused_timeout'),
+            memcache_pool_conn_get_timeout=self._conf_get(
+                'memcache_pool_conn_get_timeout'),
+            memcache_pool_socket_timeout=self._conf_get(
+                'memcache_pool_socket_timeout'))
         return token_cache
 
 
@@ -1090,6 +1311,34 @@ class _CachePool(list):
             self.append(c)
 
 
+class _MemcacheClientPool(object):
+    """An advanced memcached client pool that is eventlet safe."""
+    def __init__(self, memcache_servers, memcache_dead_retry=None,
+                 memcache_pool_maxsize=None, memcache_pool_unused_timeout=None,
+                 memcache_pool_conn_get_timeout=None,
+                 memcache_pool_socket_timeout=None):
+        # NOTE(morganfainberg): import here to avoid hard dependency on
+        # python-memcache library.
+        global _memcache_pool
+        from keystonemiddleware import _memcache_pool
+
+        self._pool = _memcache_pool.MemcacheClientPool(
+            memcache_servers,
+            arguments={
+                'dead_retry': memcache_dead_retry,
+                'socket_timeout': memcache_pool_socket_timeout,
+            },
+            maxsize=memcache_pool_maxsize,
+            unused_timeout=memcache_pool_unused_timeout,
+            conn_get_timeout=memcache_pool_conn_get_timeout,
+        )
+
+    @contextlib.contextmanager
+    def reserve(self):
+        with self._pool.get() as client:
+            yield client
+
+
 class _IdentityServer(object):
     """Operations on the Identity API server.
 
@@ -1136,7 +1385,7 @@ class _IdentityServer(object):
                       user authentication when an indeterminate
                       response is received. Optional.
         :return: token object received from keystone on success
-        :raise InvalidUserToken: if token is rejected
+        :raise InvalidToken: if token is rejected
         :raise ServiceError: if unable to authenticate token
 
         """
@@ -1180,7 +1429,7 @@ class _IdentityServer(object):
             if response.status_code == 200:
                 return data
 
-            raise InvalidUserToken()
+            raise InvalidToken()
 
     def fetch_revocation_list(self):
         try:
@@ -1356,12 +1605,22 @@ class _TokenCache(object):
 
     def __init__(self, log, cache_time=None, hash_algorithms=None,
                  env_cache_name=None, memcached_servers=None,
-                 memcache_security_strategy=None, memcache_secret_key=None):
+                 memcache_security_strategy=None, memcache_secret_key=None,
+                 use_advanced_pool=False, memcache_pool_dead_retry=None,
+                 memcache_pool_maxsize=None, memcache_pool_unused_timeout=None,
+                 memcache_pool_conn_get_timeout=None,
+                 memcache_pool_socket_timeout=None):
         self._LOG = log
         self._cache_time = cache_time
         self._hash_algorithms = hash_algorithms
         self._env_cache_name = env_cache_name
         self._memcached_servers = memcached_servers
+        self._use_advanced_pool = use_advanced_pool
+        self._memcache_pool_dead_retry = memcache_pool_dead_retry,
+        self._memcache_pool_maxsize = memcache_pool_maxsize,
+        self._memcache_pool_unused_timeout = memcache_pool_unused_timeout
+        self._memcache_pool_conn_get_timeout = memcache_pool_conn_get_timeout
+        self._memcache_pool_socket_timeout = memcache_pool_socket_timeout
 
         # memcache value treatment, ENCRYPT or MAC
         self._memcache_security_strategy = memcache_security_strategy
@@ -1375,12 +1634,36 @@ class _TokenCache(object):
 
         self._assert_valid_memcache_protection_config()
 
+    def _get_cache_pool(self, cache, memcache_servers, use_advanced_pool=False,
+                        memcache_dead_retry=None, memcache_pool_maxsize=None,
+                        memcache_pool_unused_timeout=None,
+                        memcache_pool_conn_get_timeout=None,
+                        memcache_pool_socket_timeout=None):
+        if use_advanced_pool is True and memcache_servers and cache is None:
+            return _MemcacheClientPool(
+                memcache_servers,
+                memcache_dead_retry=memcache_dead_retry,
+                memcache_pool_maxsize=memcache_pool_maxsize,
+                memcache_pool_unused_timeout=memcache_pool_unused_timeout,
+                memcache_pool_conn_get_timeout=memcache_pool_conn_get_timeout,
+                memcache_pool_socket_timeout=memcache_pool_socket_timeout)
+        else:
+            return _CachePool(cache, memcache_servers)
+
     def initialize(self, env):
         if self._initialized:
             return
 
-        self._cache_pool = _CachePool(env.get(self._env_cache_name),
-                                      self._memcached_servers)
+        self._cache_pool = self._get_cache_pool(
+            env.get(self._env_cache_name),
+            self._memcached_servers,
+            use_advanced_pool=self._use_advanced_pool,
+            memcache_dead_retry=self._memcache_pool_dead_retry,
+            memcache_pool_maxsize=self._memcache_pool_maxsize,
+            memcache_pool_unused_timeout=self._memcache_pool_unused_timeout,
+            memcache_pool_conn_get_timeout=self._memcache_pool_conn_get_timeout
+        )
+
         self._initialized = True
 
     def get(self, user_token):
@@ -1392,7 +1675,7 @@ class _TokenCache(object):
         The second element is the token data from the cache if the token was
         cached, otherwise ``None``.
 
-        :raises InvalidUserToken: if the token is invalid
+        :raises InvalidToken: if the token is invalid
 
         """
 
@@ -1443,7 +1726,7 @@ class _TokenCache(object):
     def _cache_get(self, token_id):
         """Return token information from cache.
 
-        If token is invalid raise InvalidUserToken
+        If token is invalid raise InvalidToken
         return token only if fresh (not expired).
         """
 
@@ -1492,7 +1775,7 @@ class _TokenCache(object):
         cached = jsonutils.loads(serialized)
         if cached == self._INVALID_INDICATOR:
             self._LOG.debug('Cached Token is marked unauthorized')
-            raise InvalidUserToken('Token authorization failed')
+            raise InvalidToken('Token authorization failed')
 
         data, expires = cached
 
@@ -1511,7 +1794,7 @@ class _TokenCache(object):
             return data
         else:
             self._LOG.debug('Cached Token seems expired')
-            raise InvalidUserToken('Token authorization failed')
+            raise InvalidToken('Token authorization failed')
 
     def _cache_store(self, token_id, data):
         """Store value into memcache.
